
from collections import OrderedDict

from grammar.utils import enum
from grammar import SemanticError

class IllegalType (SemanticError):
    def __init__(self, *args):
        super().__init__(*args)
        pass

class TCODE (enum):
    VALUES = ( 'VOID', 'BOOL', 'DCHAR', 'UCHAR', 'SCHAR', 'USHORT', 'SSHORT',
               'UINT', 'SINT', 'ULONG', 'SLONG', 'ULONGLONG', 'SLONGLONG',
               'FLOAT', 'DOUBLE', 'LONGDOUBLE', 'AUTO' )

class Namespace:
    def __init__(self, parent=None):
        self.parent = parent
        self.children = []
        if parent:
            self.parent.children.append(self)
        self.dict = {}
        self.anon_count = 0
        self.gen_count = 0
        if parent:
            self.key = parent.key
        else:
            self.key = ''

    def lookup_local(self, name):
        return self.dict.get(name)

    def lookup_deep(self, name):
        "instead of searching self and parents, search self and children"
        o = self.lookup_local(name)
        if o is not None: return o
        for c in self.children:
            o = c.lookup_deep(name)
            if o is not None: return o
        return None

    def lookup(self, name):
        o = self.lookup_local(name)
        if o is not None or self.parent is None: return o
        return self.parent.lookup(name)

    def anon_get_name(self):
        self.anon_count += 1
        return '+{}'.format(self.anon_count)

    def generate_name(self):
        self.gen_count += 1
        return '_generated_{}'.format(self.gen_count)

    def within(self, name):
        if self.parent is None: return False
        return self.parent.within(name)

    def full_name(self, name):
        return name

class NamedNamespace (Namespace):
    def __init__(self, parent, name):
        super().__init__(parent)
        self.name = name
        self.key = parent.key+'::'+name

    def full_name(self, name):
        r = [name]
        n = self
        while isinstance(n, NamedNamespace):
            r.append(n.name)
            n = n.parent
        r.reverse()
        return '::'.join(r)

class ParamList (Namespace):
    def __init__(self, parent):
        super().__init__(parent)
        self.dict = OrderedDict()

class Memberspace (NamedNamespace):
    def __init__(self, parent, clsname, bases):
        super().__init__(parent, clsname)
        self.clsname = clsname
        self.bases = bases
        self.dict = OrderedDict()

    def within(self, name):
        if self.clsname == name: return True
        return self.parent.within(name)

    def lookup_local(self, name):
        o = self.dict.get(name)
        if o is not None: return o
        for b in self.bases:
            o = b.lookup_local(name)
            if o is not None: return o
        return None

class NSObject:
    pass

class TypeObject:
    pass

class Typedef (NSObject, TypeObject):
    def __init__(self, name, type):
        self.name = name
        self.type = type

class TypeParam (NSObject, TypeObject):
    def __init__(self, name):
        self.name = name

class Class (NSObject, TypeObject):
    def __init__(self, name, bases, members):
        self.name = name
        self.bases = bases
        self.members = members

class TemplateClass (NSObject, TypeObject):
    def __init__(self, name, bases, members):
        self.name = name
        self.bases = bases
        self.members = members

class Enum (NSObject, TypeObject):
    pass

class NamespaceObj (NSObject):
    def __init__(self, nns):
        self.nns = nns

class NamespaceAlias (NSObject):
    def __init__(self, nns):
        self.nns = nns

class FunctionDef (NSObject):
    def __init__(self, name, type):
        self.name = name
        self.type = type

class Variable (NSObject):
    def __init__(self, name, type):
        self.name = name
        self.type = type

class Array (TypeObject):
    def __init__(self, type, dim):
        self.type = type
        self.dim = dim

class Pointer (TypeObject):
    def __init__(self, type, cv):
        self.type = type
        self.cv = cv

class Reference (TypeObject):
    def __init__(self, type):
        self.type = type

class Pointer2Member (TypeObject):
    def __init__(self, ctype, type, cv):
        self.ctype = ctype
        self.type = type
        self.cv = cv

class Function (TypeObject):
    def __init__(self, type, plist, cv):
        self.type = type
        self.plist = plist
        self.cv = cv

class CVType (TypeObject):
    def __init__(self, type, cv):
        self.type = type
        self.cv = cv

class Declare:
    def __init__(self, tns=None):
        self.tdef = False
        self.flags = set()
        self.cv = set()
        self.type = None
        self.state = 0
        self.astnode = None
        self.tns = None

    def keyword(self, kw):
        if self.state == 0:
            self.state = 1
            if kw in ('register','static','extern','mutable'):
                self.flags.add(kw)
                return
            elif kw == 'typedef':
                self.tdef = True
                self.state = 3
                return
        if self.state == 1:
            self.state = 2
            if kw in ('inline','virtual','explicit'):
                self.flags.add(kw)
                return
        if self.state == 2:
            self.state = 3
            if kw == 'friend':
                self.flags.add(kw)
                return
        if self.state == 3:
            self.state = 4
            if kw in ('const','volatile'):
                self.cv.add(kw)
                return
        if self.state == 4:
            if kw == 'auto':
                self.state = -1
                self.set_type(TCODE.AUTO)
            elif kw == 'signed':
                self.state = 5
                self.set_type(TCODE.SINT)
            elif kw == 'unsigned':
                self.state = 6
                self.set_type(TCODE.UINT)
            elif kw == 'void':
                self.state = -1
                self.set_type(TCODE.VOID)
            elif kw == 'bool':
                self.state = -1
                self.set_type(TCODE.BOOL)
            elif kw == 'float':
                self.state = -1
                self.set_type(TCODE.FLOAT)
            elif kw == 'double':
                self.state = -1
                self.set_type(TCODE.DOUBLE)
            elif kw == 'char':
                self.state = -1
                self.set_type(TCODE.DCHAR)
            elif kw == 'short':
                self.state = -1
                self.set_type(TCODE.SSHORT)
            elif kw == 'int':
                self.state = -1
                self.set_type(TCODE.SINT)
            elif kw == 'long':
                self.state = 7
                self.set_type(TCODE.SLONG)
            else:
                raise IllegalType
            return
        elif self.state == 5:
            if kw == 'char':
                self.state = -1
                self.set_type(TCODE.SCHAR, True)
            elif kw == 'short':
                self.state = -1
                self.set_type(TCODE.SSHORT, True)
            elif kw == 'int':
                self.state = -1
                self.set_type(TCODE.SINT, True)
            elif kw == 'long':
                self.state = 8
                self.set_type(TCODE.SLONG, True)
            else:
                raise IllegalType
            return
        elif self.state == 6:
            if kw == 'char':
                self.state = -1
                self.set_type(TCODE.UCHAR, True)
            elif kw == 'short':
                self.state = -1
                self.set_type(TCODE.USHORT, True)
            elif kw == 'int':
                self.state = -1
                self.set_type(TCODE.UINT, True)
            elif kw == 'long':
                self.state = 9
                self.set_type(TCODE.ULONG, True)
            else:
                raise IllegalType
            return
        elif self.state == 7:
            if kw == 'long':
                self.state = -1
                self.set_type(TCODE.SLONGLONG, True)
            elif kw == 'double':
                self.state = -1
                self.set_type(TCODE.LONGDOUBLE, True)
            else:
                raise IllegalType
            return
        elif self.state == 8:
            if kw == 'long':
                self.state = -1
                self.set_type(TCODE.SLONGLONG, True)
            else:
                raise IllegalType
            return
        elif self.state == 9:
            if kw == 'long':
                self.state = -1
                self.set_type(TCODE.ULONGLONG, True)
            else:
                raise IllegalType
            return
        else:
            raise IllegalType

    def set_type(self, type, cont=False):
        if not cont and self.type is not None:
            raise IllegalType
        if self.cv:
            self.type = CVType(type, self.cv)
        else:
            self.type = type

    def copy(self, type):
        r = Declare()
        r.tdef = self.tdef
        r.flags = self.flags
        r.type = type
        r.astnode = self.astnode
        return r

    def as_array(self, dim):
        if self.type is None:
            raise IllegalType("No type specified")
        return self.copy(Array(self.type, dim))

    def as_func(self, plist, cv):
        return self.copy(Function(self.type, plist, cv))

    def as_ref(self):
        if self.type is None:
            raise IllegalType("No type specified")
        return self.copy(Reference(self.type))

    def as_ptr(self, cv):
        if self.type is None:
            raise IllegalType("No type specified")
        return self.copy(Pointer(self.type, cv))

    def as_ptr2m(self, ct, cv):
        if self.type is None:
            raise IllegalType("No type specified")
        return self.copy(Pointer2Member(ct, self.type, cv))

    def declare(self, name, ns, ctxt, src):
        if self.tns is not None: ns = self.tns
        if self.type is None:
            raise IllegalType("No type specified", src)
        if name is None:
            name = ns.anon_get_name()
        if self.tdef:
            obj = Typedef(name, self.type)
        elif isinstance(self.type, Function):
            if self.type.type is None:
                raise IllegalType("No return type specified", src)
            obj = FunctionDef(name, self.type)
        else:
            obj = Variable(name, self.type)
        old = ns.dict.get(name)
        if old is None:
            ns.dict[name] = obj
        elif type(obj) != type(old):
            print(name, obj, old)
            raise IllegalType("Inconsistent declaration of "+name, src)
        elif not isinstance(obj, FunctionDef):
            obj = old
        else:
            ns.dict[name] = obj
        if self.astnode:
            obj.astnode = self.astnode
        return obj
